import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from ..constants import function_shemas
from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import *
from promptmodel.types.enums import ParsingType

html_output_format = """\
You must follow the provided output format. Keep the string between <> as it is.
Output format:
<response type=str>
(value here)
</response>
"""

double_bracket_output_format = """\
You must follow the provided output format. Keep the string between [[ ]] as it is.
Output format:
[[response type=str]]
(value here)
[[/response]]
"""

colon_output_format = """\
You must follow the provided output format. Keep the string before : as it is
Output format:
response type=str: (value here)

"""

tool_schemas = [{"type": "function", "function": schema} for schema in function_shemas]


def test_run_with_tools(mocker):
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    res: LLMResponse = llm.run(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
    )

    assert res.error is None, "error is not None"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "tool_calls"
    ), "finish_reason is not tool_calls"
    print(res.api_response.model_dump())
    print(res.__dict__)
    assert res.tool_calls is not None, "tool_calls is None"
    assert isinstance(res.tool_calls[0], ChatCompletionMessageToolCall)

    messages = [{"role": "user", "content": "Hello, How are you?"}]

    res: LLMResponse = llm.run(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
    )

    assert res.error is None, "error is not None"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "stop"
    ), "finish_reason is not stop"

    assert res.tool_calls is None, "tool_calls is not None"
    assert res.raw_output is not None, "raw_output is None"


@pytest.mark.asyncio
async def test_arun_with_tools(mocker):
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    res: LLMResponse = await llm.arun(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
    )

    assert res.error is None, "error is not None"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "tool_calls"
    ), "finish_reason is not tool_calls"

    assert res.tool_calls is not None, "tool_calls is None"
    assert isinstance(res.tool_calls[0], ChatCompletionMessageToolCall)

    messages = [{"role": "user", "content": "Hello, How are you?"}]

    res: LLMResponse = await llm.arun(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
    )

    assert res.error is None, "error is not None"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "stop"
    ), "finish_reason is not stop"

    assert res.tool_calls is None, "tool_calls is not None"
    assert res.raw_output is not None, "raw_output is None"


def test_run_and_parse_with_tools(mocker):
    # With parsing_type = None
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    res: LLMResponse = llm.run_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=None,
    )

    assert res.error is False, "error is not False"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "tool_calls"
    ), "finish_reason is not tool_calls"

    assert res.tool_calls is not None, "tool_calls is None"
    assert isinstance(res.tool_calls[0], ChatCompletionMessageToolCall)

    messages = [{"role": "user", "content": "Hello, How are you?"}]

    res: LLMResponse = llm.run_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=None,
    )

    assert res.error is False, "error is not False"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "stop"
    ), "finish_reason is not stop"

    assert res.tool_calls is None, "tool_calls is not None"
    assert res.raw_output is not None, "raw_output is None"

    # with parsing_type = "HTML"

    messages = [
        {
            "role": "user",
            "content": "What is the weather like in Boston? \n" + html_output_format,
        }
    ]
    llm = LLM()
    res: LLMResponse = llm.run_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )

    # 1. Output 지키고 function call ->  (Pass)
    # 2. Output 지키고 stop -> OK
    # 3. Output 무시하고 function call -> OK (function call이 나타나면 파싱을 하지 않도록 수정)

    # In this case, error is True because the output is not in the correct format
    assert res.error is False, "error is not False"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "tool_calls"
    ), "finish_reason is not tool_calls"

    assert res.tool_calls is not None, "tool_calls is None"
    assert isinstance(res.tool_calls[0], ChatCompletionMessageToolCall)

    assert res.parsed_outputs is None, "parsed_outputs is not empty"

    messages = [
        {
            "role": "user",
            "content": "Hello, How are you?\n" + html_output_format,
        }
    ]

    res: LLMResponse = llm.run_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )

    print(res.__dict__)

    if not "str" in res.raw_output:
        # if "str" in res.raw_output, it means that LLM make mistakes
        assert res.error is False, "error is not False"
        assert res.parsed_outputs is not None, "parsed_outputs is None"

    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "stop"
    ), "finish_reason is not stop"

    assert res.tool_calls is None, "tool_calls is not None"
    assert res.raw_output is not None, "raw_output is None"


@pytest.mark.asyncio
async def test_arun_and_parse_with_tools(mocker):
    # With parsing_type = None
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    res: LLMResponse = await llm.arun_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=None,
    )

    assert res.error is False, "error is not False"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "tool_calls"
    ), "finish_reason is not tool_calls"
    print(res)
    assert res.tool_calls is not None, "tool_calls is None"
    assert isinstance(res.tool_calls[0], ChatCompletionMessageToolCall)

    messages = [{"role": "user", "content": "Hello, How are you?"}]

    res: LLMResponse = await llm.arun_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=None,
    )

    assert res.error is False, "error is not False"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "stop"
    ), "finish_reason is not stop"

    assert res.tool_calls is None, "tool_calls is not None"
    assert res.raw_output is not None, "raw_output is None"

    # with parsing_type = "HTML"

    messages = [
        {
            "role": "user",
            "content": "What is the weather like in Boston? \n" + html_output_format,
        }
    ]
    llm = LLM()
    res: LLMResponse = await llm.arun_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )

    # In this case, error is False becuase if tool_calls, parsing is not performed
    assert res.error is False, "error is not False"
    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "tool_calls"
    ), "finish_reason is not tool_calls"

    assert res.tool_calls is not None, "tool_calls is None"
    assert isinstance(res.tool_calls[0], ChatCompletionMessageToolCall)

    assert res.parsed_outputs is None, "parsed_outputs is not empty"

    messages = [
        {
            "role": "user",
            "content": "Hello, How are you?\n" + html_output_format,
        }
    ]

    res: LLMResponse = await llm.arun_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    if not "str" in res.raw_output:
        assert res.error is False, "error is not False"
        assert res.parsed_outputs is not None, "parsed_outputs is None"

    assert res.api_response is not None, "api_response is None"
    assert (
        res.api_response.choices[0].finish_reason == "stop"
    ), "finish_reason is not stop"

    assert res.tool_calls is None, "tool_calls is not None"
    assert res.raw_output is not None, "raw_output is None"



from typing import (
    List,
    Dict,
    Any,
    Optional,
)
from pydantic import BaseModel
from litellm.utils import (
    ModelResponse,
    Usage,
    Message,
    Choices,
    StreamingChoices,
    Delta,
    FunctionCall,
    Function,
    ChatCompletionMessageToolCall,
)
from openai._models import BaseModel as OpenAIObject
from openai.types.chat.chat_completion import *
from openai.types.chat.chat_completion_chunk import (
    ChoiceDeltaFunctionCall,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)


class ChatLogRequest(BaseModel):
    uuid: Optional[str] = None
    message: Dict[str, Any]
    metadata: Optional[Dict] = None
    api_response: Optional[ModelResponse] = None

    def __post_init__(
        self,
    ):
        if self.api_response is not None and self.message is None:
            self.message = self.api_response.choices[0].message.model_dump()


class RunLogRequest(BaseModel):
    uuid: Optional[str] = None
    inputs: Optional[Dict[str, Any]] = None
    metadata: Optional[Dict] = None
    api_response: Optional[ModelResponse] = None



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

import nest_asyncio
from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import LLMResponse, LLMStreamResponse
from promptmodel.utils.async_utils import run_async_in_sync


def test_stream(mocker):
    llm = LLM()
    test_messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Introduce yourself in 50 words."},
    ]

    stream_res: Generator[LLMStreamResponse, None, None] = llm.stream(
        messages=test_messages,
        model="gpt-3.5-turbo",
    )
    error_count = 0
    api_responses: List[ModelResponse] = []
    for res in stream_res:
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
    print(api_responses)
    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert api_responses[0].choices[0].message.content is not None, "content is None"
    assert api_responses[0]._response_ms is not None, "response_ms is None"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    nest_asyncio.apply()
    run_async_in_sync(
        llm_proxy._async_log_to_cloud(
            log_uuid="test",
            version_uuid="test",
            inputs={},
            api_response=api_responses[0],
            parsed_outputs={},
            metadata={},
        )
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args
    api_response_dict = api_responses[0].model_dump()
    api_response_dict.update({"response_ms": api_responses[0]._response_ms})
    assert (
        kwargs["json"]["api_response"] == api_response_dict
    ), "api_response is not equal"


@pytest.mark.asyncio
async def test_astream(mocker):
    llm = LLM()
    test_messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Introduce yourself in 50 words."},
    ]

    stream_res: AsyncGenerator[LLMStreamResponse, None] = llm.astream(
        messages=test_messages,
        model="gpt-3.5-turbo",
    )
    error_count = 0
    api_responses: List[ModelResponse] = []
    async for res in stream_res:
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert api_responses[0].choices[0].message.content is not None, "content is None"
    assert api_responses[0]._response_ms is not None, "response_ms is None"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    await llm_proxy._async_log_to_cloud(
        log_uuid="test",
        version_uuid="test",
        inputs={},
        api_response=api_responses[0],
        parsed_outputs={},
        metadata={},
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args

    api_response_dict = api_responses[0].model_dump()
    api_response_dict.update({"response_ms": api_responses[0]._response_ms})

    assert (
        kwargs["json"]["api_response"] == api_response_dict
    ), "api_response is not equal"



from typing import Any, Dict
from InquirerPy import inquirer
from rich import print
from promptmodel.apis.base import APIClient


def get_org(config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Gets the current organization from the configuration.

    :return: A dictionary containing the current organization.
    """
    if "connection" not in config:
        print("User not logged in. Please run [violet]prompt login[/violet] first.")
        exit()
    if "org" not in config["connection"]:
        orgs = APIClient.execute(method="GET", path="/organizations").json()
        choices = [
            {
                "key": org["name"],
                "name": org["name"],
                "value": org,
            }
            for org in orgs
        ]
        org = inquirer.select(message="Select organization:", choices=choices).execute()
    else:
        org = config["connection"]["org"]
    return org


def get_project(config: Dict[str, Any], org: Dict[str, Any]) -> Dict[str, Any]:
    """
    Gets the current project from the configuration.

    :return: A dictionary containing the current project.
    """
    if "project" not in config["connection"]:
        projects = APIClient.execute(
            method="GET",
            path="/projects",
            params={"organization_id": org["organization_id"]},
        ).json()
        choices = [
            {
                "key": project["name"],
                "name": project["name"],
                "value": project,
            }
            for project in projects
        ]
        project = inquirer.select(message="Select project:", choices=choices).execute()
    else:
        project = config["connection"]["project"]
    return project



from __future__ import annotations

from dataclasses import dataclass
from uuid import uuid4
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    List,
    Optional,
    Coroutine,
    Union,
)
from litellm import ModelResponse

import promptmodel.utils.logger as logger
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils.async_utils import run_async_in_sync
from promptmodel.utils.config_utils import check_connection_status_decorator
from promptmodel.types.response import (
    LLMStreamResponse,
    LLMResponse,
    FunctionModelConfig,
)
from promptmodel.types.enums import InstanceType
from promptmodel.apis.base import AsyncAPIClient
from promptmodel import DevClient


@dataclass
class FunctionModelInterface:
    name: str
    default_model: str


class RegisteringMeta(type):
    def __call__(cls, *args, **kwargs):
        instance: "FunctionModel" = super().__call__(*args, **kwargs)
        # Find the global client instance in the current context
        client: Optional[DevClient] = cls.find_client_instance()
        if client is not None:
            client.register_function_model(instance.name)
        return instance

    @staticmethod
    def find_client_instance():
        import sys

        # Get the current frame (frame where FunctionModel is called)
        frame = sys._getframe(2)
        # Get global variables in the current frame
        global_vars = frame.f_globals
        # Find an instance of DevClient among global variables
        for var_name, var_val in global_vars.items():
            if isinstance(var_val, DevClient):
                return var_val
        return None


class FunctionModel(metaclass=RegisteringMeta):
    """

    Args:
        name (_type_): _description_
        version (Optional[ Union[str, int] ], optional): Choose which FunctionModel version to use. Defaults to "deploy". It can be "deploy", "latest", or version number.
        api_key (Optional[str], optional): API key for the LLM. Defaults to None. If None, use api_key in .env file.
    """

    def __init__(
        self,
        name,
        version: Optional[
            Union[str, int]
        ] = "deploy",  # "deploy" or "latest" or version number
        api_key: Optional[str] = None,
    ):
        self.name = name
        self.api_key = api_key
        self.llm_proxy = LLMProxy(name, version)
        self.version = version
        self.recent_log_uuid = None

    @check_connection_status_decorator
    def get_config(self, *args, **kwargs) -> FunctionModelConfig:
        """Get config for the promptmodel.
        It will fetch the prompt and version you specified from the Cloud. (It will be saved in cache DB, so there is no extra latency for API call.)
        - If you made A/B testing in Web Dashboard, it will fetch the prompt randomly by the A/B testing ratio.
        If dev mode is initializing, it will return None

        Returns:
            FunctionModelConfig: config for the promptmodel. It contains prompts and version_detail.
        """
        # add name to the list of function_models

        prompt, version_detail = run_async_in_sync(
            LLMProxy.fetch_prompts(self.name, self.version)
        )
        return FunctionModelConfig(
            prompts=prompt,
            model=version_detail["model"],
            name=self.name,
            version_uuid=str(version_detail["uuid"]),
            version=version_detail["version"],
            parsing_type=version_detail["parsing_type"]
            if "parsing_type" in version_detail
            else None,
            output_keys=version_detail["output_keys"]
            if "output_keys" in version_detail
            else None,
        )

    @check_connection_status_decorator
    def run(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Run FunctionModel. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find raw output in response.raw_output or response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        res: LLMResponse = self.llm_proxy.run(inputs, functions, tools, self.api_key)
        self.recent_log_uuid = res.pm_detail.log_uuid
        return res

    @check_connection_status_decorator
    async def arun(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Async run FunctionModel. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find raw output in response.raw_output or response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        res: LLMResponse = await self.llm_proxy.arun(
            inputs, functions, tools, self.api_key
        )
        self.recent_log_uuid = res.pm_detail.log_uuid
        return res

    @check_connection_status_decorator
    def stream(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Run FunctionModel with stream=True. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            Generator[LLMStreamResponse, None, None]: Generator of LLMStreamResponse. you can find raw output in response.raw_output or response.api_response['choices'][0]['delta']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        cache: Optional[LLMStreamResponse] = None
        for item in self.llm_proxy.stream(inputs, functions, tools, self.api_key):
            yield item
            cache = item

        if cache:
            self.recent_log_uuid = cache.pm_detail.log_uuid

    @check_connection_status_decorator
    async def astream(
        self,
        inputs: Optional[Dict[str, Any]] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> Coroutine[AsyncGenerator[LLMStreamResponse, None]]:
        """Async Run FunctionModel with stream=True. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            AsyncGenerator[LLMStreamResponse, None]: Generator of LLMStreamResponse. you can find raw output in response.raw_output or response.api_response['choices'][0]['delta']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """

        async def async_gen():
            cache: Optional[LLMStreamResponse] = None
            async for item in self.llm_proxy.astream(
                inputs, functions, tools, self.api_key
            ):
                yield item
                cache = item
            if cache:
                self.recent_log_uuid = cache.pm_detail.log_uuid

        return async_gen()

    @check_connection_status_decorator
    def run_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Run FunctionModel and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find parsed outputs in response.parsed_outputs. You can also find raw output in response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        res: LLMResponse = self.llm_proxy.run_and_parse(
            inputs, functions, tools, self.api_key
        )
        self.recent_log_uuid = res.pm_detail.log_uuid
        return res

    @check_connection_status_decorator
    async def arun_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Async Run FunctionModel and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find parsed outputs in response.parsed_outputs. You can also find raw output in response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        res: LLMResponse = await self.llm_proxy.arun_and_parse(
            inputs, functions, tools, self.api_key
        )
        self.recent_log_uuid = res.pm_detail.log_uuid
        return res

    @check_connection_status_decorator
    def stream_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Run FunctionModel with stream=True and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            Generator[LLMStreamResponse, None, None]: Generator of LLMStreamResponse. you can find parsed outputs in response.parsed_outputs. You can also find raw output in reaponse.raw_output.

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log
        """
        cache: Optional[LLMStreamResponse] = None
        for item in self.llm_proxy.stream_and_parse(
            inputs, functions, tools, self.api_key
        ):
            yield item
            cache = item

        if cache:
            self.recent_log_uuid = cache.pm_detail.log_uuid

    @check_connection_status_decorator
    async def astream_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> Coroutine[AsyncGenerator[LLMStreamResponse, None]]:
        """Async Run FunctionModel with stream=True and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            AsyncGenerator[LLMStreamResponse, None]: Generator of LLMStreamResponse. you can find parsed outputs in response.parsed_outputs. You can also find raw output in reaponse.raw_output.

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log
        """

        async def async_gen():
            cache: Optional[LLMStreamResponse] = None
            async for item in self.llm_proxy.astream_and_parse(
                inputs, functions, tools, self.api_key
            ):
                yield item
                cache = item
            if cache:
                self.recent_log_uuid = cache.pm_detail.log_uuid

        return async_gen()

    @check_connection_status_decorator
    async def log_score(
        self,
        log_uuid: Optional[str] = None,
        score: Optional[Dict] = {},
        *args,
        **kwargs,
    ) -> str:
        """Save Scores for RunLog of FunctionModel to the Cloud.

        Args:
            log_uuid (Optional[str], optional): UUID of the RunLog you want to save score. If None, it will use recent_log_uuid.
            score (Optional[float], optional): Scores of RunLog. Each keys will be created as Evaluation Metric in Promptmodel Dashboard. Defaults to {}.

        Returns:
            str: _description_
        """
        try:
            if log_uuid is None and self.recent_log_uuid:
                log_uuid = self.recent_log_uuid
            if log_uuid is None:
                raise Exception(
                    "log_uuid is None. Please run FunctionModel.run or FunctionModel.log first."
                )

            res = await AsyncAPIClient.execute(
                method="POST",
                path="/run_log_score",
                params={"run_log_uuid": log_uuid},
                json=score,
                use_cli_key=False,
            )
            if res.status_code != 200:
                logger.error(f"Logging error: {res}")
        except Exception as exception:
            logger.error(f"Logging error: {exception}")

    @check_connection_status_decorator
    async def log(
        self,
        version_uuid: str,
        openai_api_response: Optional[ModelResponse] = None,
        inputs: Optional[Dict[str, Any]] = {},
        parsed_outputs: Optional[Dict[str, Any]] = {},
        metadata: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ) -> str:
        """Save RunLog of FunctionModel to the Cloud.

        Args:
            version_uuid (str): version of the FunctionModel. You can find this in the FunctionModelConfig.
            openai_api_response (Optional[ModelResponse], optional): OpenAI Response Class. Defaults to None.
            inputs (Optional[Dict[str, Any]], optional): inputs dictionary for the FunctionModel. Defaults to {}.
            parsed_outputs (Optional[Dict[str, Any]], optional): parsed outputs. Defaults to {}.
            metadata (Optional[Dict[str, Any]], optional): metadata. You can save Any metadata for RunLog. Defaults to {}.

        Returns:
            str: UUID of the RunLog.
        """
        try:
            log_uuid = str(uuid4())
            content = {
                "uuid": log_uuid,
                "api_response": openai_api_response.model_dump(),
                "inputs": inputs,
                "parsed_outputs": parsed_outputs,
                "metadata": metadata,
            }

            config = kwargs["config"]
            if (
                "mask_inputs" in config["project"]
                and config["project"]["mask_inputs"] is True
            ):
                if "inputs" in content:
                    content["inputs"] = {
                        key: "PRIVATE LOGGING"
                        for key, value in content["inputs"].items()
                    }
            res = await AsyncAPIClient.execute(
                method="POST",
                path="/run_log",
                params={"version_uuid": version_uuid},
                json=content,
                use_cli_key=False,
            )
            if res.status_code != 200:
                logger.error(f"Logging error: {res}")

            self.recent_log_uuid = log_uuid
            return log_uuid
        except Exception as exception:
            logger.error(f"Logging error: {exception}")


class PromptModel(FunctionModel):
    """Deprecated"""

    def __init__(self, name, version: Optional[Union[str, int]] = "deploy"):
        super().__init__(name, version)



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any
from litellm import ModelResponse

from promptmodel.apis.base import APIClient, AsyncAPIClient
from promptmodel.utils.crypto import decrypt_message, generate_api_key, encrypt_message


@pytest.mark.asyncio
async def test_get_headers(mocker):
    api_client = APIClient()
    async_api_client = AsyncAPIClient()

    mock_exit = mocker.patch("builtins.exit", side_effect=SystemExit)
    api_key = generate_api_key()
    mock_envirion_get = MagicMock(return_value=api_key)

    gt_cli_output = {
        "Authorization": f"Bearer {decrypt_message(encrypt_message(api_key))}"
    }
    gt_api_output = {"Authorization": f"Bearer {api_key}"}
    # No user
    mocker.patch("promptmodel.apis.base.read_config", return_value={})
    with pytest.raises(SystemExit):
        res = api_client._get_headers()
    mock_exit.assert_called_once()
    mock_exit.reset_mock()

    with pytest.raises(SystemExit):
        res = await async_api_client._get_headers()
    mock_exit.assert_called_once()
    mock_exit.reset_mock()

    mocker.patch("promptmodel.apis.base.read_config", return_value={"user": {}})

    with pytest.raises(Exception):
        res = api_client._get_headers()
    with pytest.raises(Exception):
        res = await async_api_client._get_headers()

    mocker.patch(
        "promptmodel.apis.base.read_config",
        return_value={"user": {"encrypted_api_key": encrypt_message(api_key)}},
    )
    res = api_client._get_headers()
    assert res == gt_cli_output, "API key is not decrypted properly"
    res = await async_api_client._get_headers()
    assert res == gt_api_output, "API key is not decrypted properly"

    mocker.patch("promptmodel.apis.base.os.environ.get", mock_envirion_get)
    res = api_client._get_headers(use_cli_key=False)
    assert res == gt_api_output, "API key is not retrieved properly"
    res = await async_api_client._get_headers(use_cli_key=False)
    assert res == gt_api_output, "API key is not retrieved properly"


@pytest.mark.asyncio
async def test_execute(mocker):
    api_client = APIClient()
    async_api_client = AsyncAPIClient()
    mock_exit = mocker.patch("builtins.exit", side_effect=SystemExit)
    mocker.patch("promptmodel.apis.base.APIClient._get_headers", return_value={})
    mocker.patch("promptmodel.apis.base.AsyncAPIClient._get_headers", return_value={})

    mock_request = mocker.patch(
        "promptmodel.apis.base.requests.request", return_value=None
    )
    mock_async_request = mocker.patch(
        "promptmodel.apis.base.httpx.AsyncClient.request", return_value=None
    )
    with pytest.raises(SystemExit):
        res = api_client.execute(path="test")
    mock_request.assert_called_once()
    mock_request.reset_mock()
    mock_exit.assert_called_once()
    mock_exit.reset_mock()

    res = await async_api_client.execute(path="test")
    mock_async_request.assert_called_once()
    mock_async_request.reset_mock()

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_request = mocker.patch(
        "promptmodel.apis.base.requests.request", return_value=mock_response
    )
    mock_async_request = mocker.patch(
        "promptmodel.apis.base.httpx.AsyncClient.request", return_value=mock_response
    )

    res = api_client.execute(path="test")
    mock_request.assert_called_once()
    assert res == mock_response, "Response is not returned properly"
    mock_request.reset_mock()

    res = await async_api_client.execute(path="test")
    mock_async_request.assert_called_once()
    assert res == mock_response, "Response is not returned properly"
    mock_async_request.reset_mock()

    mock_response.status_code = 403
    mock_request = mocker.patch(
        "promptmodel.apis.base.requests.request", return_value=mock_response
    )
    mock_async_request = mocker.patch(
        "promptmodel.apis.base.httpx.AsyncClient.request", return_value=mock_response
    )
    with pytest.raises(SystemExit):
        res = api_client.execute(path="test")
    mock_request.assert_called_once()
    mock_request.reset_mock()
    mock_exit.assert_called_once()
    mock_exit.reset_mock()

    res = await async_api_client.execute(path="test")
    mock_async_request.assert_called_once()
    mock_async_request.reset_mock()

    mock_response.status_code = 500
    mock_request = mocker.patch(
        "promptmodel.apis.base.requests.request", return_value=mock_response
    )
    mock_async_request = mocker.patch(
        "promptmodel.apis.base.httpx.AsyncClient.request", return_value=mock_response
    )
    with pytest.raises(SystemExit):
        res = api_client.execute(path="test")
    mock_request.assert_called_once()
    mock_request.reset_mock()
    mock_exit.assert_called_once()
    mock_exit.reset_mock()

    res = await async_api_client.execute(path="test")
    mock_async_request.assert_called_once()
    mock_async_request.reset_mock()



import pytest
import asyncio
import nest_asyncio
from typing import Coroutine
from unittest.mock import AsyncMock

from promptmodel.utils.async_utils import run_async_in_sync

nest_asyncio.apply()


def test_sync_context(mocker):
    coro = AsyncMock(return_value="test")
    res = run_async_in_sync(coro())
    coro.assert_called_once()

    assert res == "test", "res is not test"


def test_sync_async_context(mocker):
    coro = AsyncMock(return_value="test")

    async def async_context(coro: Coroutine):
        res = await coro
        return res

    res = asyncio.run(async_context(coro()))
    coro.assert_called_once()

    assert res == "test", "res is not test"


@pytest.mark.asyncio
async def test_async_context(mocker):
    coro = AsyncMock(return_value="test")
    res = await coro()
    coro.assert_called_once()

    assert res == "test", "res is not test"


@pytest.mark.asyncio
async def test_async_sync_context(mocker):
    coro = AsyncMock(return_value="test")
    print("ready")

    def sync_context(coro: Coroutine):
        return run_async_in_sync(coro)

    res = sync_context(coro())
    coro.assert_called_once()

    assert res == "test", "res is not test"



import time
import asyncio
import typer
import importlib
import signal
from typing import Dict, Any, List
from playhouse.shortcuts import model_to_dict

import webbrowser
from rich import print
from InquirerPy import inquirer
from watchdog.observers import Observer

from promptmodel import DevApp
from promptmodel.apis.base import APIClient
from promptmodel.constants import ENDPOINT_URL, WEB_CLIENT_URL
from promptmodel.cli.commands.init import init as promptmodel_init
from promptmodel.cli.utils import get_org, get_project
from promptmodel.cli.signal_handler import dev_terminate_signal_handler
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.websocket import DevWebsocketClient, CodeReloadHandler
from promptmodel.database.orm import initialize_db


def connect():
    """Connect websocket and opens up DevApp in the browser."""
    upsert_config({"initializing": True}, "connection")
    signal.signal(signal.SIGINT, dev_terminate_signal_handler)
    promptmodel_init(from_cli=False)

    config = read_config()

    if "project" not in config["connection"]:
        org = get_org(config)
        project = get_project(config=config, org=org)

        # connect
        res = APIClient.execute(
            method="POST",
            path="/project/cli_connect",
            params={"project_uuid": project["uuid"]},
        )
        if res.status_code != 200:
            print(f"Error: {res.json()['detail']}")
            return

        upsert_config(
            {
                "project": project,
                "org": org,
            },
            section="connection",
        )

    else:
        org = config["connection"]["org"]
        project = config["connection"]["project"]

        res = APIClient.execute(
            method="POST",
            path="/project/cli_connect",
            params={"project_uuid": project["uuid"]},
        )
        if res.status_code != 200:
            print(f"Error: {res.json()['detail']}")
            return

    _devapp_filename, devapp_instance_name = "promptmodel_dev:app".split(":")

    devapp_module = importlib.import_module(_devapp_filename)
    devapp_instance: DevApp = getattr(devapp_module, devapp_instance_name)

    dev_url = f"{WEB_CLIENT_URL}/org/{org['slug']}/projects/{project['uuid']}"

    # Open websocket connection to backend server
    dev_websocket_client = DevWebsocketClient(_devapp=devapp_instance)

    import threading

    try:
        main_loop = asyncio.get_running_loop()
    except RuntimeError:
        main_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(main_loop)

    # save samples, FunctionSchema, FunctionModel, ChatModel to cloud server in dev_websocket_client.connect_to_gateway
    res = APIClient.execute(
        method="POST",
        path="/save_instances_in_code",
        params={"project_uuid": project["uuid"]},
        json={
            "function_models": devapp_instance._get_function_model_name_list(),
            "chat_models": devapp_instance._get_chat_model_name_list(),
            "function_schemas": devapp_instance._get_function_schema_list(),
            "samples": devapp_instance.samples,
        },
    )
    if res.status_code != 200:
        print(f"Error: {res.json()['detail']}")
        return

    print(
        f"\nOpening [violet]Promptmodel[/violet] prompt engineering environment with the following configuration:\n"
    )
    print(f"📌 Organization: [blue]{org['name']}[/blue]")
    print(f"📌 Project: [blue]{project['name']}[/blue]")
    print(
        f"\nIf browser doesn't open automatically, please visit [link={dev_url}]{dev_url}[/link]"
    )
    webbrowser.open(dev_url)

    upsert_config({"online": True, "initializing": False}, section="connection")

    reloader_thread = threading.Thread(
        target=start_code_reloader,
        args=(_devapp_filename, devapp_instance_name, dev_websocket_client, main_loop),
    )
    reloader_thread.daemon = True  # Set the thread as a daemon
    reloader_thread.start()

    # Open Websocket
    asyncio.run(
        dev_websocket_client.connect_to_gateway(
            project_uuid=project["uuid"],
            connection_name=project["name"],
            cli_access_header=APIClient._get_headers(),
        )
    )


app = typer.Typer(invoke_without_command=True, callback=connect)


def start_code_reloader(
    _devapp_filename, devapp_instance_name, dev_websocket_client, main_loop
):
    event_handler = CodeReloadHandler(
        _devapp_filename, devapp_instance_name, dev_websocket_client, main_loop
    )
    observer = Observer()
    observer.schedule(event_handler, path=".", recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



"""
Prompt & model versioning on the cloud, built for developers.
"""
from setuptools import setup, find_namespace_packages

# Read README.md for the long description
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="promptmodel",
    version="0.1.13",
    packages=find_namespace_packages(),
    entry_points={
        "console_scripts": [
            "prompt = promptmodel.cli.main:app",
        ],
    },
    description="Prompt & model versioning on the cloud, built for developers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="weavel",
    url="https://github.com/weavel-ai/promptmodel",
    install_requires=[
        "httpx[http2]",
        "pydantic>=2.4.2",
        "peewee",
        "typer[all]",
        "cryptography",
        "pyyaml",
        "InquirerPy",
        "litellm>=1.7.1",
        # "litellm@git+https://github.com/weavel-ai/litellm.git@llms_add_clova_support",
        "python-dotenv",
        "websockets",
        "termcolor",
        "watchdog",
        "readerwriterlock",
        "nest-asyncio",
    ],
    python_requires=">=3.8.10",
    keywords=[
        "weavel",
        "agent",
        "llm",
        "tools",
        "promptmodel",
        "llm agent",
        "prompt",
        "versioning",
        "eval",
        "evaluation",
        "collaborative",
    ],
)



import time
from promptmodel.apis.base import APIClient
import typer
import webbrowser
from rich import print

from promptmodel.constants import ENDPOINT_URL, GRANT_ACCESS_URL
from promptmodel.utils.config_utils import upsert_config
from promptmodel.utils.crypto import generate_api_key, encrypt_message


def login():
    """Authenticate Client CLI."""
    # TODO: Check if already logged in
    api_key = generate_api_key()
    encrypted_key = encrypt_message(api_key)
    upsert_config({"encrypted_api_key": encrypted_key}, section="connection")
    url = f"{GRANT_ACCESS_URL}?token={api_key}"
    webbrowser.open(url)
    print("Please grant access to the CLI by visiting the URL in your browser.")
    print("Once you have granted access, you can close the browser tab.")
    print(f"\nURL: [link={url}]{url}[/link]\n")
    print("Waiting...\n")
    waiting_time = 0
    while waiting_time < 300:
        # Check access every 5 seconds
        try:
            res = APIClient.execute("/cli_access/check", ignore_auth_error=True)
            if res.json() == True:
                print("[green]Access granted![/green] 🎉")
                print(
                    "Run [violet][bold]prompt init[/bold][/violet] to start developing prompts.\n"
                )
                return
        except Exception as err:
            print(f"[red]Error: {err}[/red]")
        time.sleep(5)
        waiting_time += 5
    print("Please try again later.")


app = typer.Typer(invoke_without_command=True, callback=login)



import random


def select_version_by_ratio(versions):
    epsilon = 1e-10
    ratios = [version["ratio"] for version in versions]

    if not abs(sum(ratios) - 1.0) <= epsilon:
        raise ValueError(f"Sum of ratios must be 1.0, now {sum(ratios)}")

    cumulative_ratios = []
    cumulative_sum = 0
    for ratio in ratios:
        cumulative_sum += ratio
        cumulative_ratios.append(cumulative_sum)

    random_value = random.random()
    for idx, cumulative_ratio in enumerate(cumulative_ratios):
        if random_value <= cumulative_ratio:
            return versions[idx]



from typing import (
    List,
    Dict,
    Any,
    Optional,
)
from pydantic import BaseModel
from litellm.utils import (
    ModelResponse,
    Usage,
    Message,
    Choices,
    StreamingChoices,
    Delta,
    FunctionCall,
    Function,
    ChatCompletionMessageToolCall,
)
from openai._models import BaseModel as OpenAIObject
from openai.types.chat.chat_completion import *
from openai.types.chat.chat_completion_chunk import (
    ChoiceDeltaFunctionCall,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)


class PMDetail(BaseModel):
    model: str
    name: str
    version_uuid: str
    version: int
    log_uuid: str


class LLMResponse(OpenAIObject):
    api_response: Optional[ModelResponse] = None
    raw_output: Optional[str] = None
    parsed_outputs: Optional[Dict[str, Any]] = None
    error: Optional[bool] = None
    error_log: Optional[str] = None
    function_call: Optional[FunctionCall] = None
    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None
    pm_detail: Optional[PMDetail] = None


class LLMStreamResponse(OpenAIObject):
    api_response: Optional[ModelResponse] = None
    raw_output: Optional[str] = None
    parsed_outputs: Optional[Dict[str, Any]] = None
    error: Optional[bool] = None
    error_log: Optional[str] = None
    function_call: Optional[ChoiceDeltaFunctionCall] = None
    tool_calls: Optional[List[ChoiceDeltaToolCall]] = None
    pm_detail: Optional[PMDetail] = None


class FunctionModelConfig(BaseModel):
    """Response Class for FunctionModel.get_config()
    prompts: List[Dict[str, Any]] = []
        each prompt can have role, content, name, function_call, and tool_calls
    version_detail: Dict[str, Any] = {}
        version_detail has "model", "uuid", "parsing_type" and "output_keys".
    model: str
        model name (e.g. "gpt-3.5-turbo")
    name: str
        name of the FunctionModel.
    version_uuid: str
        version uuid of the FunctionModel.
    version: int
        version id of the FunctionModel.
    parsing_type: Optional[str] = None
        parsing type of the FunctionModel.
    output_keys: Optional[List[str]] = None
        output keys of the FunctionModel.
    """

    prompts: List[Dict[str, Any]]
    model: str
    name: str
    version_uuid: str
    version: int
    parsing_type: Optional[str] = None
    output_keys: Optional[List[str]] = None


class PromptModelConfig(FunctionModelConfig):
    """Deprecated. Use FunctionModelConfig instead."""


class ChatModelConfig(BaseModel):
    system_prompt: str
    model: str
    name: str
    version_uuid: str
    version: int
    message_logs: Optional[List[Dict]] = []


class FunctionSchema(BaseModel):
    """
    {
            "name": str,
            "description": Optional[str],
            "parameters": {
                "type": "object",
                "properties": {
                    "argument_name": {
                        "type": str,
                        "description": Optional[str],
                        "enum": Optional[List[str]]
                    },
                },
                "required": Optional[List[str]],
            },
        }
    """

    class _Parameters(BaseModel):
        class _Properties(BaseModel):
            type: str
            description: Optional[str] = ""
            enum: Optional[List[str]] = []

        type: str = "object"
        properties: Dict[str, _Properties] = {}
        required: Optional[List[str]] = []

    name: str
    description: Optional[str] = None
    parameters: _Parameters



from __future__ import annotations
import dis
import nest_asyncio
from dataclasses import dataclass
from typing import Callable, Dict, Any, List, Optional, Union

from promptmodel.types.response import FunctionSchema


@dataclass
class FunctionModelInterface:
    name: str


@dataclass
class ChatModelInterface:
    name: str


class DevClient:
    """DevClient main class"""

    def __init__(self):
        self.function_models: List[FunctionModelInterface] = []
        self.chat_models: List[ChatModelInterface] = []

    def register(self, func):
        instructions = list(dis.get_instructions(func))
        for idx in range(
            len(instructions) - 1
        ):  # We check up to len-1 because we access idx+1 inside loop
            instruction = instructions[idx]
            # print(instruction)
            if instruction.opname in ["LOAD_ATTR", "LOAD_METHOD", "LOAD_GLOBAL"] and (
                instruction.argval == "FunctionModel"
                or instruction.argval == "ChatModel"
            ):
                next_instruction = instructions[idx + 1]

                # Check if the next instruction is LOAD_CONST with string value
                if next_instruction.opname == "LOAD_CONST" and isinstance(
                    next_instruction.argval, str
                ):
                    if instruction.argval == "FunctionModel":
                        self.function_models.append(
                            FunctionModelInterface(name=next_instruction.argval)
                        )
                    elif instruction.argval == "ChatModel":
                        self.chat_models.append(
                            ChatModelInterface(name=next_instruction.argval)
                        )

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def register_function_model(self, name):
        for function_model in self.function_models:
            if function_model.name == name:
                return

        self.function_models.append(FunctionModelInterface(name=name))

    def register_chat_model(self, name):
        for chat_model in self.chat_models:
            if chat_model.name == name:
                return

        self.chat_models.append(ChatModelInterface(name=name))

    def _get_function_model_name_list(self) -> List[str]:
        return [function_model.name for function_model in self.function_models]


class DevApp:
    _nest_asyncio_applied = False

    def __init__(self):
        self.function_models: List[FunctionModelInterface] = []
        self.chat_models: List[ChatModelInterface] = []
        self.samples: List[Dict[str, Any]] = []
        self.functions: Dict[
            str, Dict[str, Union[FunctionSchema, Optional[Callable]]]
        ] = {}

        if not DevApp._nest_asyncio_applied:
            DevApp._nest_asyncio_applied = True
            nest_asyncio.apply()

    def include_client(self, client: DevClient):
        self.function_models.extend(client.function_models)
        self.chat_models.extend(client.chat_models)

    def register_function(
        self, schema: Union[Dict[str, Any], FunctionSchema], function: Callable
    ):
        function_name = schema["name"]
        if isinstance(schema, dict):
            try:
                schema = FunctionSchema(**schema)
            except:
                raise ValueError("schema is not a valid function call schema.")

        if function_name not in self.functions:
            self.functions[function_name] = {
                "schema": schema,
                "function": function,
            }

    def _call_register_function(self, name: str, arguments: Dict[str, str]):
        function_to_call: Optional[Callable] = self.functions[name]["function"]
        if not function_to_call:
            return
        try:
            function_response = function_to_call(**arguments)
            return function_response
        except Exception as e:
            raise e

    def _get_function_name_list(self) -> List[str]:
        return list(self.functions.keys())

    def _get_function_schema_list(self) -> List[Dict]:
        return [
            self.functions[function_name]["schema"].model_dump()
            for function_name in self._get_function_name_list()
        ]

    def _get_function_schemas(self, function_names: List[str] = []):
        try:
            function_schemas = [
                self.functions[function_name]["schema"].model_dump()
                for function_name in function_names
            ]
            return function_schemas
        except Exception as e:
            raise e

    def register_sample(self, name: str, content: Dict[str, Any]):
        self.samples.append({"name": name, "content": content})

    def _get_function_model_name_list(self) -> List[str]:
        return [function_model.name for function_model in self.function_models]

    def _get_chat_model_name_list(self) -> List[str]:
        return [chat_model.name for chat_model in self.chat_models]



from .enums import InstanceType
from .response import (
    LLMResponse,
    LLMStreamResponse,
    FunctionModelConfig,
    PromptModelConfig,
    ChatModelConfig,
    FunctionSchema,
)
from .request import ChatLogRequest, RunLogRequest



"""Logger module"""

import os
from typing import Any
import termcolor


def debug(msg: Any, *args):
    if os.environ.get("TESTMODE_LOGGING", "false") != "true":
        return
    print(termcolor.colored("[DEBUG] " + str(msg) + str(*args), "light_yellow"))


def success(msg: Any, *args):
    if os.environ.get("TESTMODE_LOGGING", "false") != "true":
        return
    print(termcolor.colored("[SUCCESS] " + str(msg) + str(*args), "green"))


def info(msg: Any, *args):
    if os.environ.get("TESTMODE_LOGGING", "false") != "true":
        return
    print(termcolor.colored("[INFO] " + str(msg) + str(*args), "blue"))


def warning(msg: Any, *args):
    if os.environ.get("TESTMODE_LOGGING", "false") != "true":
        return
    print(termcolor.colored("[WARNING] " + str(msg) + str(*args), "yellow"))


def error(msg: Any, *args):
    print(termcolor.colored("[Error] " + str(msg) + str(*args), "red"))



from typing import Dict, List, Optional, Tuple, Any
from uuid import UUID
from playhouse.shortcuts import model_to_dict
from peewee import fn, JOIN

from promptmodel.utils import logger
from promptmodel.database.config import db


# def delete_fake_sessions():
#     """Delete ChatLogSession which len(ChatLog) == 1 where ChatLog.session_uuid == ChatLogSession.uuid"""
#     with db.atomic():
#         sessions_to_delete: List[ChatLogSession] = list(
#             ChatLogSession.select()
#             .join(ChatLog, JOIN.LEFT_OUTER)
#             .group_by(ChatLogSession)
#             .having(fn.COUNT(ChatLog.id) <= 1)
#         )
#         (
#             (
#                 ChatLogSession.delete().where(
#                     ChatLogSession.uuid.in_(
#                         [session.uuid for session in sessions_to_delete]
#                     )
#                 )
#             ).execute()
#         )
#     return


# def hide_chat_model_not_in_code(local_chat_model_list: List):
#     return (
#         ChatModel.update(used_in_code=False)
#         .where(ChatModel.name.not_in(local_chat_model_list))
#         .execute()
#     )


# def update_chat_model_uuid(local_uuid, new_uuid):
#     """Update ChatModel.uuid"""
#     if str(local_uuid) == str(new_uuid):
#         return
#     else:
#         with db.atomic():
#             local_chat_model: ChatModel = ChatModel.get(ChatModel.uuid == local_uuid)
#             ChatModel.create(
#                 uuid=new_uuid,
#                 name=local_chat_model.name,
#                 project_uuid=local_chat_model.project_uuid,
#                 created_at=local_chat_model.created_at,
#                 used_in_code=local_chat_model.used_in_code,
#                 is_deployed=True,
#             )
#             ChatModelVersion.update(function_model_uuid=new_uuid).where(
#                 ChatModelVersion.chat_model_uuid == local_uuid
#             ).execute()
#             ChatModel.delete().where(ChatModel.uuid == local_uuid).execute()
#         return


# def update_candidate_chat_model_version(new_candidates: dict):
#     """Update candidate ChatModelVersion's candidate_version_id(version)"""
#     with db.atomic():
#         for uuid, version in new_candidates.items():
#             (
#                 ChatModelVersion.update(version=version, is_deployed=True)
#                 .where(ChatModelVersion.uuid == uuid)
#                 .execute()
#             )
#         # Find ChatModel
#         chat_model_versions: List[ChatModelVersion] = list(
#             ChatModelVersion.select().where(
#                 ChatModelVersion.uuid.in_(list(new_candidates.keys()))
#             )
#         )
#         chat_model_uuids = [
#             chat_model.chat_model_uuid.uuid for chat_model in chat_model_versions
#         ]
#         ChatModel.update(is_deployed=True).where(
#             ChatModel.uuid.in_(chat_model_uuids)
#         ).execute()


# def rename_chat_model(chat_model_uuid: str, new_name: str):
#     """Update the name of the given ChatModel."""
#     return (
#         ChatModel.update(name=new_name)
#         .where(ChatModel.uuid == chat_model_uuid)
#         .execute()
#     )


# def fetch_chat_log_with_uuid(session_uuid: str):
#     try:
#         try:
#             chat_logs: List[ChatLog] = list(
#                 ChatLog.select()
#                 .where(ChatLog.session_uuid == UUID(session_uuid))
#                 .order_by(ChatLog.created_at.asc())
#             )
#         except:
#             return []

#         chat_log_to_return = [
#             {
#                 "role": chat_log.role,
#                 "content": chat_log.content,
#                 "tool_calls": chat_log.tool_calls,
#             }
#             for chat_log in chat_logs
#         ]
#         return chat_log_to_return
#     except Exception as e:
#         logger.error(e)
#         return None


# def get_latest_version_chat_model(
#     chat_model_name: str,
#     session_uuid: Optional[str] = None,
# ) -> Tuple[List[Dict[str, Any]], str]:
#     try:
#         if session_uuid:
#             if type(session_uuid) == str:
#                 session_uuid = UUID(session_uuid)
#             session: ChatLogSession = ChatLogSession.get(
#                 ChatLogSession.uuid == session_uuid
#             )

#             version: ChatModelVersion = ChatModelVersion.get(
#                 ChatModelVersion.uuid == session.version_uuid
#             )

#             instruction: List[Dict[str, Any]] = [version["system_prompt"]]
#         else:
#             with db.atomic():
#                 try:
#                     sessions_with_version: List[ChatLogSession] = list(
#                         ChatLogSession.select()
#                         .join(ChatModelVersion)
#                         .where(
#                             ChatModelVersion.chat_model_uuid
#                             == ChatModel.get(ChatModel.name == chat_model_name).uuid
#                         )
#                     )
#                     session_uuids = [x.uuid for x in sessions_with_version]

#                     latest_chat_log: List[ChatLog] = list(
#                         ChatLog.select()
#                         .where(ChatLog.session_uuid.in_(session_uuids))
#                         .order_by(ChatLog.created_at.desc())
#                     )

#                     latest_chat_log: ChatLog = latest_chat_log[0]
#                     latest_session: ChatLogSession = ChatLogSession.get(
#                         ChatLogSession.uuid == latest_chat_log.session_uuid
#                     )

#                     version: ChatModelVersion = (
#                         ChatModelVersion.select()
#                         .where(ChatModelVersion.uuid == latest_session.uuid)
#                         .get()
#                     )
#                 except:
#                     version: ChatModelVersion = list(
#                         ChatModelVersion.select()
#                         .join(ChatModel)
#                         .where(ChatModel.name == chat_model_name)
#                         .order_by(ChatModelVersion.created_at.desc())
#                         .get()
#                     )

#                 instruction: List[Dict[str, Any]] = [version["system_prompt"]]

#             version_details = {
#                 "model": version.model,
#                 "uuid": version.uuid,
#             }

#             return instruction, version_details

#     except Exception as e:
#         logger.error(e)
#         return None, None


# def find_ancestor_chat_model_version(
#     chat_model_version_uuid: str, versions: Optional[list] = None
# ):
#     """Find ancestor ChatModel version"""

#     # get all versions
#     if versions is None:
#         response = list(ChatModelVersion.select())
#         versions = [model_to_dict(x, recurse=False) for x in response]

#     # find target version
#     target = list(
#         filter(lambda version: version["uuid"] == chat_model_version_uuid, versions)
#     )[0]

#     target = _find_ancestor(target, versions)

#     return target


# def find_ancestor_chat_model_versions(target_chat_model_uuid: Optional[str] = None):
#     """find ancestor versions for each versions in input"""
#     # get all versions
#     if target_chat_model_uuid is not None:
#         response = list(
#             ChatModelVersion.select().where(
#                 ChatModelVersion.chat_model_uuid == target_chat_model_uuid
#             )
#         )
#     else:
#         response = list(ChatModelVersion.select())
#     versions = [model_to_dict(x, recurse=False) for x in response]

#     targets = list(
#         filter(
#             lambda version: version["status"] == ModelVersionStatus.CANDIDATE.value
#             and version["version"] is None,
#             versions,
#         )
#     )

#     targets = [
#         find_ancestor_chat_model_version(target["uuid"], versions) for target in targets
#     ]
#     targets_with_real_ancestor = [target for target in targets]

#     return targets_with_real_ancestor


# def _find_ancestor(target: dict, versions: List[Dict]):
#     ancestor = None
#     temp = target
#     if target["from_uuid"] is None:
#         ancestor = None
#     else:
#         while temp["from_uuid"] is not None:
#             new_temp = [
#                 version for version in versions if version["uuid"] == temp["from_uuid"]
#             ][0]
#             if (
#                 new_temp["version"] is not None
#                 or new_temp["status"] == ModelVersionStatus.CANDIDATE.value
#             ):
#                 ancestor = new_temp
#                 break
#             else:
#                 temp = new_temp
#         target["from_uuid"] = ancestor["uuid"] if ancestor is not None else None

#     return target



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from promptmodel.llms.llm import LLM, ParseResult
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import LLMResponse, LLMStreamResponse
from promptmodel.types.enums import ParsingType
from promptmodel import FunctionModel, DevClient
from promptmodel.dev_app import FunctionModelInterface
from promptmodel.function_model import RegisteringMeta

client = DevClient()


def test_find_client(mocker):
    pm = FunctionModel("test")
    assert client.function_models == [FunctionModelInterface(name="test")]


def test_get_config(mocker, mock_fetch_prompts):
    fetch_prompts = mocker.patch(
        "promptmodel.function_model.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    # mock registering_meta
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")
    assert len(client.function_models) == 1
    config = function_model.get_config()
    assert len(config.prompts) == 2


def test_run(mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")
    res: LLMResponse = function_model.run({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = function_model.run()
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


@pytest.mark.asyncio
async def test_arun(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")

    res: LLMResponse = await function_model.arun({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = await function_model.arun()
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


def test_stream(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")

    res: Generator[LLMStreamResponse, None, None] = function_model.stream({})
    chunks: List[LLMStreamResponse] = []
    for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = function_model.stream()
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


@pytest.mark.asyncio
async def test_astream(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")

    res: AsyncGenerator[LLMStreamResponse, None] = await function_model.astream({})
    chunks: List[LLMStreamResponse] = []
    async for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = await function_model.astream({})
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


def test_run_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")
    res: LLMResponse = function_model.run_and_parse({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs == {}

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = function_model.run_and_parse({})
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


@pytest.mark.asyncio
async def test_arun_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")

    res: LLMResponse = await function_model.arun_and_parse({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs == {}

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = await function_model.arun_and_parse({})
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


def test_stream_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")

    res: Generator[LLMStreamResponse, None, None] = function_model.stream_and_parse({})
    chunks: List[LLMStreamResponse] = []
    for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = function_model.stream_and_parse({})
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()


@pytest.mark.asyncio
async def test_astream_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )
    mocker.patch("promptmodel.function_model.RegisteringMeta", MagicMock())
    function_model = FunctionModel("test")

    res: AsyncGenerator[
        LLMStreamResponse, None
    ] = await function_model.astream_and_parse({})
    chunks: List[LLMStreamResponse] = []
    async for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res = await function_model.astream_and_parse({})
    print(res)
    fetch_prompts.assert_not_called()
    async_log_to_cloud.assert_not_called()



import os
import sys
import typer
from promptmodel.cli.commands.login import app as login
from promptmodel.cli.commands.init import app as init

# from promptmodel.cli.commands.dev import app as dev
from promptmodel.cli.commands.connect import app as connect
from promptmodel.cli.commands.project import app as project
from promptmodel.cli.commands.configure import app as configure
from promptmodel.cli.commands.fix import app as fix

# 현재 작업 디렉토리를 sys.path에 추가
current_working_directory = os.getcwd()
if current_working_directory not in sys.path:
    sys.path.append(current_working_directory)

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)

app.add_typer(login, name="login")
app.add_typer(init, name="init")
# app.add_typer(dev, name="dev")
app.add_typer(connect, name="connect")
app.add_typer(project, name="project")
app.add_typer(configure, name="configure")
app.add_typer(fix, name="fix")

if __name__ == "__main__":
    app()
